import os
import sys
import urllib.request
from tqdm import tqdm
import zipfile
import json


def downloadurltofile(url, filename):
    if not os.path.exists(filename):
        print(f"--> Downloading {filename} <--".center(80, "#"))
        with open(filename, "wb") as file:
            with urllib.request.urlopen(url) as resp:
                length = int(resp.getheader("content-length"))
                blocksize = max(4096, length // 100)
                with tqdm(total=length, file=sys.stdout) as pbar:
                    while True:
                        buff = resp.read(blocksize)
                        if not buff:
                            break
                        file.write(buff)
                        pbar.update(len(buff))
        print(" Download complete ".center(80, "#"))
    else:
        print(f"-->> {filename} file already exists locally <<--".center(80, "#"))
    print()


def download(url, targetfolder, targetfile):
    path = os.getcwd()
    data_dir = os.path.abspath(os.path.join(path, targetfolder))

    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    targetfile = os.path.join(data_dir, targetfile)
    downloadurltofile(url, targetfile)
    return data_dir, targetfile


def unzippedfile(folder, file):
    with zipfile.ZipFile(file, "r") as zip_ref:
        zip_ref.extractall(folder)


def upload_data(workspace, datastore, src_dir, tgt_path):
    datastore.upload(src_dir=src_dir, target_path=tgt_path, show_progress=True)
    print(" Upload complete ".center(80, "#"))


def get_config(configfile):
    jsonfile = open(configfile)
    configdata = json.load(jsonfile)
    return configdata



# Imports
import os
import sys
import uuid
import time
import socket
import argparse
import threading
import subprocess
import logging

####---from mpi4py import MPI


def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


def flush(proc, proc_log):
    while True:
        proc_out = proc.stdout.readline()
        if proc_out == "" and proc.poll() is not None:
            proc_log.close()
            break
        elif proc_out:
            sys.stdout.write(proc_out)
            proc_log.write(proc_out)
            proc_log.flush()


def evaluate_cmd(cmd, logFileName):
    cmd_log = open(logFileName, "a")
    cmd_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    cmd_flush = threading.Thread(target=flush, args=(cmd_proc, cmd_log))
    cmd_flush.start()
    flush(cmd_proc, cmd_log)
    return cmd_proc


if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    install("azureml-sdk")
    install("notebook")
    from azureml.core import Run
    from notebook.notebookapp import list_running_servers

    ### PARSE ARGUMENTS
    parser = argparse.ArgumentParser()
    parser.add_argument("--jupyter_token", default=uuid.uuid1().hex)
    parser.add_argument("--jupyter_port", default=8888)
    parser.add_argument("--use_gpu", default=False)
    parser.add_argument("--n_gpus_per_node", default=0)

    args, unparsed = parser.parse_known_args()

    ### CONFIGURE GPU RUN
    GPU_run = args.use_gpu

    if GPU_run:
        n_gpus_per_node = eval(args.n_gpus_per_node)

    ip = socket.gethostbyname(socket.gethostname())

    data = {
        "jupyter": ip + ":" + str(args.jupyter_port),
        "token": args.jupyter_token,
    }

    jupyter = data["jupyter"]
    token = data["token"]

    logger.debug("- args: ", args)
    logger.debug("- unparsed: ", unparsed)
    logger.debug("- my ip is ", ip)

    ####---# if rank == 0:
    running_jupyter_servers = list(list_running_servers())
    logger.debug("- running jupyter servers", running_jupyter_servers)

    ### if any jupyter processes running
    ### KILL'EM ALL!!!
    if len(running_jupyter_servers) > 0:
        for server in running_jupyter_servers:
            os.system(f'kill {server["pid"]}')

    ### RECORD LOGS
    run = Run.get_context()

    run.log("jupyter", jupyter)
    run.log("token", token)

    workspace_name = run.experiment.workspace.name.lower()
    run_id = run.get_details()["runId"]

    mount_point = f"/mnt/batch/tasks/shared/LS_root/jobs/{workspace_name}/azureml/{run_id.lower()}/mounts/"

    cmd = (
        f" jupyter lab --ip 0.0.0.0 --port {args.jupyter_port}"
        f" --NotebookApp.token={token}"
    )
    cmd += f" --notebook-dir={mount_point}"
    cmd += f" --allow-root --no-browser"

    jupyter_log = open("jupyter_log.txt", "a")
    jupyter_proc = subprocess.Popen(
        cmd.split(),
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    jupyter_flush = threading.Thread(target=flush, args=(jupyter_proc, jupyter_log))
    jupyter_flush.start()

    while not list(list_running_servers()):
        time.sleep(5)

    jupyter_servers = list(list_running_servers())
    assert len(jupyter_servers) == 1, "more than one jupyter server is running"

    flush(jupyter_proc, jupyter_log)

    extraCMDS = {
        f"wget https://api.ngc.nvidia.com/v2/resources/nvidia/med/getting_started/versions/1/zip": "unload_log.txt",
        f"mv zip zip.zip": "mv_log.txt",
        f"unzip zip.zip": "unzip_log.txt",
    }
    extraCMDS_procs = {}

    for key in extraCMDS:
        extraCMDS_procs[key] = evaluate_cmd(key, extraCMDS[key])

    if jupyter_proc:
        jupyter_proc.kill()

    for key in extraCMDS_procs:
        if extraCMDS_procs[key]:
            extraCMDS_procs[key].kill()

    run.complete()
    run.cancel()



import logging

from azureml.core import Workspace, Environment
from azureml.core.conda_dependencies import CondaDependencies
from azureml.core.compute import ComputeTarget, AmlCompute

import click, os

from azureml_ngc_tools.AzureMLComputeCluster import AzureMLComputeCluster
from azureml_ngc_tools.cli import ngccontent
from azureml.exceptions._azureml_exception import ProjectSystemException

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--login", is_flag=False, required=True, help="Path to the login config file"
)
@click.option("--app", is_flag=False, required=True, help="Path to the config file")
@click.version_option()
def start(login, app):
    login_config = ngccontent.get_config(login)
    app_config = ngccontent.get_config(app)

    ### WORKSPACE
    subscription_id = login_config["azureml_user"]["subscription_id"]
    resource_group = login_config["azureml_user"]["resource_group"]
    workspace_name = login_config["azureml_user"]["workspace_name"]

    try:
        ws = Workspace(
            workspace_name=workspace_name,
            subscription_id=subscription_id,
            resource_group=resource_group,
        )
    except ProjectSystemException:
        msg = f'\n\nThe workspace "{workspace_name}" does not exist. '
        msg += f'Go to \n\n  -->> https://docs.microsoft.com/en-us/azure/machine-learning/how-to-manage-workspace <<-- \n\n'
        msg += f'and create the workspace first.\n\n\n'
        msg += f'Your current configuration: \n\n'
        msg += f'Workspace name: {workspace_name} \n'
        msg += f'Subscription id: {subscription_id} \n'
        msg += f'Resource group: {resource_group}\n\n'
        raise Exception(msg)

    verify = f"""
    Subscription ID: {subscription_id}
    Resource Group: {resource_group}
    Workspace: {workspace_name}"""
    print(verify)

    ### experiment name
    exp_name = login_config["aml_compute"]["exp_name"]

    ### azure ml names
    ct_name = login_config["aml_compute"]["ct_name"]
    vm_name = login_config["aml_compute"]["vm_name"].lower()
    vm_priority = login_config["aml_compute"]["vm_priority"]

    ### trust but verify
    verify = f"""
    Experiment name: {exp_name}"""
    print(verify)

    ### GPU RUN INFO
    workspace_vm_sizes = AmlCompute.supported_vmsizes(ws)
    workspace_vm_sizes = [
        (e["name"].lower(), e["gpus"])
        for e in workspace_vm_sizes
        if "_nc" in e["name"].lower()
    ]
    workspace_vm_sizes = dict(workspace_vm_sizes)

    if vm_name in workspace_vm_sizes:
        # vm_name = login_config["aml_compute"]["vm_name"]
        gpus_per_node = workspace_vm_sizes[vm_name]

        verify = f"""
    Compute target: {ct_name}
    VM Size: {vm_name}
    No of GPUs: {gpus_per_node}
    Priority: {vm_priority}
        """
        print(verify)

        ### get SSH keys
        ssh_key_pub, pri_key_file = get_ssh_keys()

        if ct_name not in ws.compute_targets:
            print(f"Compute target {ct_name} does not exist...")
            config = AmlCompute.provisioning_configuration(
                vm_size=vm_name,
                min_nodes=login_config["aml_compute"]["min_nodes"],
                max_nodes=login_config["aml_compute"]["max_nodes"],
                vm_priority=vm_priority,
                idle_seconds_before_scaledown=login_config["aml_compute"][
                    "idle_seconds_before_scaledown"
                ],
                admin_username=login_config["aml_compute"]["admin_name"],
                admin_user_ssh_key=ssh_key_pub,
                remote_login_port_public_access="Enabled",
            )
            ct = ComputeTarget.create(ws, ct_name, config)
            ct.wait_for_completion(show_output=True)
        else:
            print(
                "    Loading pre-existing compute target {ct_name}".format(
                    ct_name=ct_name
                )
            )
            ct = ws.compute_targets[ct_name]
    else:
        print("Unsupported vm_size {vm_size}".format(vm_size=vm_name))
        print("The specified vm size must be one of ...")
        for azure_gpu_vm_size in workspace_vm_sizes.keys():
            print("... " + azure_gpu_vm_size)
        raise Exception(
            "{vm_size} does not have Pascal or above GPU Family".format(vm_size=vm_name)
        )

    environment_name = login_config["aml_compute"]["environment_name"]
    python_interpreter = login_config["aml_compute"]["python_interpreter"]
    conda_packages = login_config["aml_compute"]["conda_packages"]

    if environment_name not in ws.environments:
        env = Environment(name=environment_name)
        env.docker.enabled = login_config["aml_compute"]["docker_enabled"]
        env.docker.base_image = None
        env.docker.base_dockerfile = f'FROM {app_config["base_dockerfile"]}'
        env.python.interpreter_path = python_interpreter
        env.python.user_managed_dependencies = True
        conda_dep = CondaDependencies()

        for conda_package in conda_packages:
            conda_dep.add_conda_package(conda_package)

        env.python.conda_dependencies = conda_dep
        env.register(workspace=ws)
        evn = env
    else:
        env = ws.environments[environment_name]

    amlcluster = AzureMLComputeCluster(
        workspace=ws,
        compute_target=ct,
        initial_node_count=1,
        experiment_name=login_config["aml_compute"]["exp_name"],
        environment_definition=env,
        jupyter_port=login_config["aml_compute"]["jupyter_port"],
        telemetry_opt_out=True,
        admin_username=login_config["aml_compute"]["admin_name"],
        admin_ssh_key=pri_key_file,
    )

    print(f"\n    Go to: {amlcluster.jupyter_link}")
    print("    Press Ctrl+C to stop the cluster.")


def get_ssh_keys():
    from cryptography.hazmat.primitives import serialization as crypto_serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from cryptography.hazmat.backends import default_backend as crypto_default_backend

    dir_path = os.path.join(os.getcwd(), ".ssh")

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    pub_key_file = os.path.join(dir_path, "key.pub")
    pri_key_file = os.path.join(dir_path, "key")

    keys_exist = True

    if not os.path.exists(pub_key_file):
        print("Public SSH key does not exist!")
        keys_exist = False

    if not os.path.exists(pri_key_file):
        print("Private SSH key does not exist!")
        keys_exist = False

    if not keys_exist:
        key = rsa.generate_private_key(
            backend=crypto_default_backend(), public_exponent=65537, key_size=2048
        )

        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption(),
        )
        public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )

        with open(pub_key_file, "wb") as f:
            f.write(public_key)

        with open(pri_key_file, "wb") as f:
            f.write(private_key)

        os.chmod(pri_key_file, 0o600)

    with open(pub_key_file, "r") as f:
        pubkey = f.read()

    return pubkey, pri_key_file


def go():
    start()


if __name__ == "__main__":
    go()



from . import AzureMLComputeCluster

__all__ = ["AzureMLComputeCluster"]

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python

from os.path import exists
from setuptools import setup, find_packages

import versioneer

setup(
    name="azureml_ngc_tools",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="AzureML integration with NGC",
    url="",
    keywords="azureml,ngc,gpu",
    license="BSD",
    packages=find_packages(),
    include_package_data=True,
    long_description_content_type="text/markdown",
    long_description=(open("README.md").read() if exists("README.md") else ""),
    zip_safe=False,
    install_requires=list(open("requirements.txt").read().strip().split("\n")),
    entry_points="""
    [console_scripts]
    azureml-ngc-tools=azureml_ngc_tools.cli.azureml_ngc:go
    """,
    python_requires=">=3.5",
)



#!/usr/bin/env python3
from azureml.core import Experiment
from azureml.core.compute import AmlCompute
from azureml.train.estimator import Estimator

import time, os, subprocess, logging
import pathlib
import threading
import signal

logger = logging.getLogger(__name__)


class AzureMLComputeCluster:
    """ Deploy a Dask cluster using Azure ML

    This creates a dask scheduler and workers on an Azure ML Compute Target.

    Parameters
    ----------
    workspace: azureml.core.Workspace (required)
        Azure ML Workspace - see https://aka.ms/azureml/workspace

    compute_target: azureml.core.ComputeTarget (required)
        Azure ML Compute Target - see https://aka.ms/azureml/computetarget

    environment_definition: azureml.core.Environment (required)
        Azure ML Environment - see https://aka.ms/azureml/environments

    experiment_name: str (optional)
        The name of the Azure ML Experiment used to control the cluster.

        Defaults to ``dask-cloudprovider``.

    initial_node_count: int (optional)
        The initial number of nodes for the Dask Cluster.

        Defaults to ``1``.

    jupyter_port: int (optional)
        Port on headnode to use for hosting JupyterLab session.

        Defaults to ``9000``.

    additional_ports: list[tuple[int, int]] (optional)
        Additional ports to forward. This requires a list of tuples where the first element
        is the port to open on the headnode while the second element is the port to map to
        or forward via the SSH-tunnel.

        Defaults to ``[]``.

    admin_username: str (optional)
        Username of the admin account for the AzureML Compute.
        Required for runs that are not on the same VNET. Defaults to empty string.
        Throws Exception if machine not on the same VNET.

        Defaults to ``""``.

    admin_ssh_key: str (optional)
        Location of the SSH secret key used when creating the AzureML Compute.
        The key should be passwordless if run from a Jupyter notebook.
        The ``id_rsa`` file needs to have 0700 permissions set.
        Required for runs that are not on the same VNET. Defaults to empty string.
        Throws Exception if machine not on the same VNET.

        Defaults to ``""``.

    telemetry_opt_out: bool (optional)
        A boolean parameter. Defaults to logging a version of AzureMLCluster
        with Microsoft. Set this flag to False if you do not want to share this
        information with Microsoft. Microsoft is not tracking anything else you
        do in your Dask cluster nor any other information related to your
        workload.

    asynchronous: bool (optional)
        Flag to run jobs asynchronously.

    **kwargs: dict
        Additional keyword arguments.
    """

    def __init__(
        self,
        workspace,
        compute_target,
        environment_definition,
        experiment_name=None,
        initial_node_count=None,
        jupyter=None,
        jupyter_port=None,
        additional_ports=None,
        admin_username=None,
        admin_ssh_key=None,
        telemetry_opt_out=None,
        **kwargs,
    ):
        ### REQUIRED PARAMETERS
        self.workspace = workspace
        self.compute_target = compute_target
        self.environment_definition = environment_definition

        ### EXPERIMENT DEFINITION
        self.experiment_name = experiment_name
        self.tags = {"tag": "azureml-ngc-tools"}

        ### ENVIRONMENT AND VARIABLES
        self.initial_node_count = initial_node_count

        ### SEND TELEMETRY
        self.telemetry_opt_out = telemetry_opt_out
        self.telemetry_set = False

        ### GPU RUN INFO
        self.workspace_vm_sizes = AmlCompute.supported_vmsizes(self.workspace)
        self.workspace_vm_sizes = [
            (e["name"].lower(), e["gpus"]) for e in self.workspace_vm_sizes
        ]
        self.workspace_vm_sizes = dict(self.workspace_vm_sizes)

        self.compute_target_vm_size = self.compute_target.serialize()["properties"][
            "status"
        ]["vmSize"].lower()
        self.n_gpus_per_node = self.workspace_vm_sizes[self.compute_target_vm_size]
        self.use_gpu = True if self.n_gpus_per_node > 0 else False

        ### JUPYTER AND PORT FORWARDING
        self.jupyter = jupyter
        self.jupyter_port = jupyter_port
        self.portforward_proc = None
        self.end_logging = False  # FLAG FOR STOPPING THE port_forward_logger THREAD

        if additional_ports is not None:
            if type(additional_ports) != list:
                error_message = (
                    f"The additional_ports parameter is of {type(additional_ports)}"
                    " type but needs to be a list of int tuples."
                    " Check the documentation."
                )
                logger.exception(error_message)
                raise TypeError(error_message)

            if len(additional_ports) > 0:
                if type(additional_ports[0]) != tuple:
                    error_message = (
                        f"The additional_ports elements are of {type(additional_ports[0])}"
                        " type but needs to be a list of int tuples."
                        " Check the documentation."
                    )
                    raise TypeError(error_message)

                ### check if all elements are tuples of length two and int type
                all_correct = True
                for el in additional_ports:
                    if type(el) != tuple or len(el) != 2:
                        all_correct = False
                        break

                    if (type(el[0]), type(el[1])) != (int, int):
                        all_correct = False
                        break

                if not all_correct:
                    error_message = (
                        f"At least one of the elements of the additional_ports parameter"
                        " is wrong. Make sure it is a list of int tuples."
                        " Check the documentation."
                    )
                    raise TypeError(error_message)

        self.additional_ports = [] if additional_ports is None else additional_ports

        self.admin_username = admin_username
        self.admin_ssh_key = admin_ssh_key

        ### FUTURE EXTENSIONS
        self.kwargs = kwargs

        ### ABSOLUTE PATH
        self.abs_path = pathlib.Path(__file__).parent.absolute()

        ### close the cluster handler
        signal.signal(signal.SIGINT, self.__signal_handler)

        ### define script parameters
        self.script_params = {}
        self.script_params["--use_gpu"] = self.use_gpu
        self.script_params["--n_gpus_per_node"] = self.n_gpus_per_node

        ### headnode info
        self.headnode_info = {}

        if not self.telemetry_opt_out:
            self.__append_telemetry()

        self.__create_cluster()
        self.__print_message("Cluster created...")

    def __signal_handler(self, signal, frame):
        print()
        self.__print_message("Closing the cluster...")
        self._close()
        # sys.exit(0)

    def __append_telemetry(self):
        if not self.telemetry_set:
            self.telemetry_set = True
            try:
                from azureml._base_sdk_common.user_agent import append

                append("AzureML-NGC-Tools", "0.1")
            except ImportError:
                pass

    def __print_message(self, msg, length=80, filler="#", pre_post=""):
        logger.info(msg)
        print(f"{pre_post} {msg} {pre_post}".center(length, filler))

    def __create_cluster(self):
        print("\n")
        self.__print_message("Setting up cluster")

        exp = Experiment(self.workspace, self.experiment_name)
        estimator = Estimator(
            os.path.join(self.abs_path, "setup"),
            compute_target=self.compute_target,
            entry_script="start_jupyter.py",
            environment_definition=self.environment_definition,
            script_params=self.script_params,
            node_count=1,  ### start only scheduler
            use_docker=True,
        )
        run = exp.submit(estimator, tags=self.tags)
        self.run = run
        self.status = "running"

        self.__print_message("Waiting for compute cluster's IP")
        while (
            run.get_status() != "Canceled"
            and run.get_status() != "Failed"
            and "jupyter"
            not in run.get_metrics()  # and "scheduler" not in run.get_metrics()
        ):
            print(".", end="")
            logger.info("Compute Cluster not ready")
            time.sleep(5)

        if run.get_status() == "Canceled" or run.get_status() == "Failed":
            logger.exception("Failed to start the AzureML Compute Cluster")
            raise Exception("Failed to start the AzureML Compute Cluster.")

        print()
        self.__print_message("Jupyter session is running...")
        print("\n\n")

        self.__setup_port_forwarding()
        self.__update_links()

        self.__print_message("Connections established")

    def __update_links(self):
        hostname = "localhost"
        location = self.workspace.get_details()["location"]
        token = self.run.get_metrics()["token"]

        self.headnode_info[
            "jupyter_url"
        ] = f"http://{hostname}:{self.jupyter_port}/?token={token}"

        logger.info(f'Jupyter URL:   {self.headnode_info["jupyter_url"]}')

    def __port_forward_logger(self, portforward_proc):
        portforward_log = open("portforward_out_log.txt", "w")

        while True:
            portforward_out = portforward_proc.stdout.readline()
            if portforward_proc != "":
                portforward_log.write(portforward_out)
                portforward_log.flush()

            if self.end_logging:
                break
        return

    def __setup_port_forwarding(self):
        jupyter_address = self.run.get_metrics()["jupyter"]
        headnode_ip = self.run.get_metrics()["jupyter"].split(":")[0]

        headnode_public_ip = self.compute_target.list_nodes()[0]["publicIpAddress"]
        headnode_public_port = self.compute_target.list_nodes()[0]["port"]
        self.__print_message("headnode_public_ip: {}".format(headnode_public_ip))
        self.__print_message("headnode_public_port: {}".format(headnode_public_port))

        cmd = (
            "ssh -vvv -o StrictHostKeyChecking=no -N"
            f" -i {os.path.expanduser(self.admin_ssh_key)}"
            f" -L 0.0.0.0:{self.jupyter_port}:{headnode_ip}:8888"
        )

        cmd += f" {self.admin_username}@{headnode_public_ip} -p {headnode_public_port}"

        self.portforward_proc = subprocess.Popen(
            cmd.split(),
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        ### Starting thread to keep the SSH tunnel open on Windows
        portforward_logg = threading.Thread(
            target=self.__port_forward_logger, args=[self.portforward_proc]
        )
        portforward_logg.start()

    @property
    def jupyter_link(self):
        """ Link to JupyterLab on running on the headnode of the cluster.
        Set ``jupyter=True`` when creating the ``AzureMLCluster``.
        """
        try:
            link = self.headnode_info["jupyter_url"]
        except KeyError:
            return ""
        else:
            return link

    # close cluster
    def _close(self):
        if self.status == "closed":
            return

        if self.run:
            self.run.complete()
            self.run.cancel()

        self.status = "closed"
        self.__print_message("Cluster has now been closed.")

        if self.portforward_proc is not None:
            ### STOP LOGGING SSH
            self.portforward_proc.terminate()
            self.end_logging = True

    def close(self):
        """ Close the cluster. All Azure ML Runs corresponding to the scheduler
        and worker processes will be completed. The Azure ML Compute Target will
        return to its minimum number of nodes after its idle time before scaledown.
        """
        return self._close()
